-- List1.lua: List functions 1
-- 2021-Feb-09 20.32 / TNA

local pairs = pairs
local type = type

_ENV = nil -- Ensure no accidental globals

local List1 = {}

List1.progName = "List1"
List1.progStamp = "2021-May-07 11.23"

function List1.map( f, l )

  local m = {}

  for k, v in pairs( l ) do
    m[k] = f( l[k] )
  end

  return m

end

function List1.filter( f, l )

  local m = {}

  for k, v in pairs( l ) do
    if f( l[k] ) then
      m[k] = l[k]
    end
  end

  return m

end

function List1.foldl( f, z, l )

  local r = z

  for k, v in pairs( l ) do
    r = f( r, l[k] )
  end

  return r

end

function List1.product( l )

  return List1.foldl( function ( x, y ) return x*y end, 1, l )

end

function List1.numericIndexLimits( l )

  local minIndex
  local maxIndex

  for k in pairs( l ) do
    if type( k ) == "number" then
      if not minIndex or k < minIndex then
        minIndex = k
      end
      if not maxIndex or k > maxIndex then
        maxIndex = k
      end
    end
  end

  return minIndex, maxIndex

end

function List1.borders( l )

  local bs = {}
  local _, maxX = List1.numericIndexLimits( l )

  local border = 0

  for i = 1, (maxX or 0) do
    if l[i] ~= nil then
      if border + 1 < i then
        bs[#bs+1] = border
      end
      border = i
    end
  end

  bs[#bs+1] = border

  return bs

end

function List1.isProperSequence( l )

  local minX, _ = List1.numericIndexLimits( l )

  return (not minX or 1) == 1 and #List1.borders( l ) == 1

end

return List1
