-- Sieve1.lua: Prime sieve 1
-- 2021-Feb-09 16.57 / TNA

local print = print
local pairs = pairs
local math = math
local io = io
local List = require "List1"
local Integer = require "Integer1"
-- local ll = require "LuaLibrary"

_ENV = nil -- Ensure no accidental globals

local Sieve1 = {}

Sieve1.progName = "Sieve1"
Sieve1.progStamp = "2021-May-07 12.23"

-- print( Sieve1.progName .. "(" .. Sieve1.progStamp .. "): ...=<", ..., ">" )

function Sieve1.leastFactor( n )

  for d = 2, n do
    if n % d == 0 then
      return d
    end
    if n // d <= d then
      return n, d
    end
  end

  return 1

end

function Sieve1.isPrime( n )
  return Sieve1.leastFactor( n ) == n
end

function Sieve1.primesN( n )

  local ps = {}
  local p = 1

  for i = 1, n do
    repeat
      p = p + 1
    until Sieve1.isPrime( p )
    ps[i] = p
  end

  return ps

end

Sieve1.moduleCount = 3

Sieve1.modules = Sieve1.primesN( Sieve1.moduleCount )

Sieve1.moduleProduct = List.product( Sieve1.modules )

Sieve1.isWableRemainder = {}
Sieve1.toWheelRemainder = {}
Sieve1.fromWheelRemainder = {}
Sieve1.wableRemainderCount = 0
for i = 0, Sieve1.moduleProduct - 1 do
  Sieve1.isWableRemainder[i] = Integer.gcd( i, Sieve1.moduleProduct ) == 1
  if Sieve1.isWableRemainder[i] then
    Sieve1.toWheelRemainder[i] = Sieve1.wableRemainderCount
    Sieve1.fromWheelRemainder[Sieve1.wableRemainderCount] = i
    Sieve1.wableRemainderCount = Sieve1.wableRemainderCount + 1
  end
end

function Sieve1.isWable( x )
  return Sieve1.isWableRemainder[ x % Sieve1.moduleProduct ]
end

function Sieve1.toWheel( n )
  return ( n // Sieve1.moduleProduct ) * Sieve1.wableRemainderCount + Sieve1.toWheelRemainder[n % Sieve1.moduleProduct]
end

function Sieve1.fromWheel( k )
  return ( k // Sieve1.wableRemainderCount ) * Sieve1.moduleProduct + Sieve1.fromWheelRemainder[k % Sieve1.wableRemainderCount]
end

function Sieve1.integerBits()

  local v = math.mininteger
  local bits = 0

  while v ~= 0 do
    v = v >> 1
    bits = bits + 1
  end

  return bits

end

Sieve1.bits = Sieve1.integerBits()

Sieve1.maskStep = Sieve1.wableRemainderCount // Integer.gcd( Sieve1.wableRemainderCount, Sieve1.bits )

function Sieve1.vNext( vFirst, vDelta )
  while not Sieve1.isWable( vFirst ) do
    vFirst = vFirst + vDelta
  end
  return Sieve1.toWheel( vFirst ), vFirst
end

function Sieve1.maskNext( v, w, vDelta )

  local index = w // Sieve1.bits
  local mask = 0

  repeat
    mask = mask | ( 1 << ( w % Sieve1.bits ) )
    repeat
      v = v + vDelta
    until Sieve1.isWable( v )
    w = Sieve1.toWheel( v )
  until w // Sieve1.bits ~= index

  return index, mask, v, w

end

function Sieve1.toTable0( t, index, mask, maxValue )
  for i = 0, Sieve1.bits - 1 do
    if mask & ( 1 << i ) == 0 then
      local v = Sieve1.fromWheel( index*Sieve1.bits + i )
      if maxValue and v > maxValue then
        return t
      end
      t[#t+1] = v
    end
  end
  return t
end

function Sieve1.toTable( index, mask, maxValue )
  return Sieve1.toTable0( {}, index, mask, maxValue )
end

function Sieve1.allTable( list, inxMax, maxValue )
  local t = {}
  for i = 1, Sieve1.moduleCount do
    if maxValue and Sieve1.modules[i] > maxValue then
      return t
    end
    t[#t+1] = Sieve1.modules[i]
  end
  local fix = 1
  for i = 0, inxMax do
    t = Sieve1.toTable0( t, i, list[i] | fix, maxValue )
    fix = 0
  end
  return t
end

function Sieve1.primeNext( pw, list )
  repeat
    pw = pw + 1
  until list[pw // Sieve1.bits] & ( 1 << ( pw % Sieve1.bits ) ) == 0
  return pw
end

function Sieve1.doSieve( pw, list, inxMax )
  local p = Sieve1.fromWheel( pw )
-- io.write( p )
  local inx, mask, v, w = Sieve1.maskNext( p*p, Sieve1.toWheel( p*p ), p )
-- print( "doSieve: mask=" .. ll.toString( Sieve1.toTable( inx, mask ) ) .. ", v=" .. v .. ", w=" .. w )
  if inx > inxMax then
    return 0
  end
  list[inx] = (list[inx] or 0) | mask
  local c = 1
  local step = p * Sieve1.maskStep
  for n = 1, step do
    inx, mask, v, w = Sieve1.maskNext( v, w, p )
    if inx > inxMax then
      return c
    end
    for i = inx, inxMax, step do
-- print( "doSieve(p=" .. p .. "): n=" .. n .. ", i=" .. i )
      list[i] = (list[i] or 0) | mask
      c = c + 1
    end
  end
  return c
end

function Sieve1.allSieve( inxMax )
  local list = {[0] = 0}
  local pw = 0
  repeat
    pw = Sieve1.primeNext( pw, list )
-- print( "allSieve: fromWheel(pw=" .. pw .. ") = " .. Sieve1.fromWheel( pw ) )
    local c = Sieve1.doSieve( pw, list, inxMax )
-- io.write( "/", c, " " )
  until c == 0
  return list
end

function Sieve1.pi( x )
  if x <= 1 then
    return 0
  end
  local c = 0
  for i = 1, Sieve1.moduleCount do
    if Sieve1.modules[i] <= x then
      c = c + 1
    end
  end
  local wx, vx = Sieve1.vNext( x, -1 )
  local inxMax = wx // Sieve1.bits
-- print( "pi0(x=" .. x .. "): inxMax=" .. inxMax )
  local list = Sieve1.allSieve( inxMax )
  local tLast
-- print( "pi1(x=" .. x .. "): c=" .. c )
  for i = 0, inxMax do
    tLast = Sieve1.toTable( i, list[i], x )
    c = c + #tLast
-- print( "pi2(x=" .. x .. "): i=" .. i .. ", c=" .. c )
  end
  return c - 1
end

function Sieve1.primesTo( x )
  local intMax = Sieve1.vNext( x, -1 ) // Sieve1.bits
  return Sieve1.allTable( Sieve1.allSieve( intMax ), intMax, x )
end

return Sieve1
